
#
# ------------------------------------------------------------------------------

from .__version__ import *



__name__ = 'BaseExtensions'
__author__ = "Tyler Stegmaier"
__email__ = "tyler.stegmaier.510@gmail.com"
__copyright__ = "Copyright 2020"
__credits__ = [
        "Copyright (c) 2020 Tyler Stegmaier",
        ]
__license__ = "MIT"
__version__ = version
__maintainer__ = __author__
__maintainer_email__ = __email__

# How mature is this project? Common values are
#   3 - Alpha
#   4 - Beta
#   5 - Production/Stable
__status__ = 'Development Status :: 3 - Alpha'

__url__ = r'https://github.com/Jakar510/BaseExtensions'
# download_url=f'https://github.com/Jakar510/PyDebug/TkinterExtensions/releases/tag/{version}'
__classifiers__ = [
        __status__,

        # Indicate who your project is intended for
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',

        # Pick your license as you wish
        'License :: Free To Use But Restricted',

        # Support platforms
        'Operating System :: MacOS',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX',

        'Programming Language :: Python :: 3',
        ]


__short_description__ = ''



import os
from typing import overload

from PIL import Image




__all__ = [
        'ResizePhoto',
        ]
def ResizePhoto(image: Image.Image, *, WidthMax: int or float, HeightMax: int or float) -> Image:
    scalingFactor = min((WidthMax / image.width, HeightMax / image.height))
    newSize = (int(scalingFactor * image.width), int(scalingFactor * image.height))
    return image.resize(newSize)



@overload
def IsImage(path: str) -> bool: ...
@overload
def IsImage(directory: str, fileName: str) -> bool: ...

def IsImage(path: str = None, directory: str = None, fileName: str = None, ) -> bool:
    try:
        if directory and fileName: path = os.path.join(directory, fileName)

        assert (os.path.isfile(path))
        with open(path, 'rb') as f:
            with Image.open(f) as img:
                assert (isinstance(img, Image.Image))
                img.verify()
                return True

    except (FileNotFoundError, ValueError, EOFError, Image.UnidentifiedImageError, Image.DecompressionBombError):
        return False



import copy
from enum import Enum
from json import dumps, loads

from .MixIns import *




__all__ = [
        'BaseModel', 'BaseListModel', 'BaseDictModel', 'BaseSetModel'
        ]

_T = TypeVar("_T")
_KT = TypeVar("_KT")
_VT = TypeVar("_VT")


class BaseModel(object):
    @staticmethod
    def Assert(t: type, d: object):
        if not isinstance(d, t): BaseModel.throw(d, t)

    @staticmethod
    @overload
    def throw(d: object, t: Type): ...
    @staticmethod
    @overload
    def throw(d: object, *types: Type): ...
    @staticmethod
    def throw(d: object, *types: Type):
        if not types: raise ValueError('expected types must be provided')

        if len(types) == 1: raise TypeError(f'Expecting {types[0]}   got type {type(d)}')

        raise TypeError(f'Expecting one of {types}   got type {type(d)}')

    @staticmethod
    def AssertKeys(d: dict, *args):
        try:
            for key in args: assert (key in d)
        except AssertionError:
            raise KeyError(f"Required keys {args} are missing.     existing keys: {d.keys()}")

    @staticmethod
    def ConvertBool(o: Union[bool, str]) -> bool:
        if isinstance(o, bool): return o
        if isinstance(o, str): return o.lower() == 'true'

        BaseModel.throw(o, bool)

    @staticmethod
    def RaiseKeyError(key, d: Dict): raise KeyError(f'{key} not in {d.keys()}')



    def Clone(self): return copy.deepcopy(self)
    def ToString(self) -> str:
        try:
            return f'<{self.__class__.__qualname__} Object() [ {self.ToJsonString()} ]'
        except AttributeError:
            return f'<{self.__class__.__name__} Object() [ {self.ToJsonString()} ]'
    def _Filter(self, func: callable): raise NotImplementedError()
    def ToDict(self): raise NotImplementedError()
    def ToList(self): raise NotImplementedError()
    def enumerate(self): raise NotImplementedError()
    def Count(self) -> int: raise NotImplementedError()
    def Empty(self) -> bool: raise NotImplementedError()


    @classmethod
    def Parse(cls, d): return cls()
    @classmethod
    def FromJson(cls, string: Union[str, bytes, bytearray], **kwargs): return cls.Parse(loads(string, **kwargs))
    def ToJsonString(self) -> str: return dumps(self, indent=4, default=self.serialize)  # , cls=JsonEncoder)
    @staticmethod
    def serialize(o):
        if isinstance(o, Enum): return o.value

        if isinstance(o, BaseSetModel): return o.ToList()

        if isinstance(o, BaseListModel): return o

        if isinstance(o, BaseDictModel): return o.ToDict()

        if hasattr(o, 'ToList') and callable(o.ToList): return o.ToList()

        if hasattr(o, 'ToDict') and callable(o.ToDict): return o.ToDict()

        return o

class BaseListModel(list, BaseModel, List[_T]):
    def __init__(self, source: Union[List, Iterable] = None):
        super().__init__(source or [])
    def __str__(self): return self.ToString()

    def __contains__(self, item: _T): return super().__contains__(item)
    def __setitem__(self, key: int, value: _T): return super().__setitem__(key, value)
    def __getitem__(self, key: int) -> Optional[_T]:
        try: return super().__getitem__(key)
        except KeyError: return None
    def __iter__(self) -> Iterable[_T]: return super().__iter__()
    def enumerate(self) -> Iterable[Tuple[int, _T]]: return enumerate(self)

    @property
    def Count(self) -> int: return len(self)
    @property
    def Empty(self) -> bool: return self.Count == 0

    def Iter(self) -> Iterable[int]: return range(len(self))

    def _Filter(self, func: callable) -> List[_T]: return list(filter(func, self))
    def ToList(self) -> List[_T]: return list(self)


    def ToDict(self) -> Dict[int, _T]:
        d = { }
        for key, value in self.enumerate():
            if hasattr(value, 'ToDict') and callable(value.ToDict):
                d[key] = value.ToDict()
            else:
                d[key] = value
        return d


    @classmethod
    def Parse(cls, d):
        if isinstance(d, list):
            return cls(d)

        BaseModel.throw(d, list)

class BaseSetModel(set, BaseModel, Set[_T]):
    def __str__(self): return self.ToString()
    def __contains__(self, item: _T): return super().__contains__(item)
    def __delitem__(self, item: _T): return self.discard(item)
    # def __isub__(self, item: _T): return self.discard(item)
    def __iadd__(self, item: _T): return self.add(item)
    def __iter__(self) -> Iterable[_T]: return super().__iter__()
    def enumerate(self) -> Iterable[Tuple[int, _T]]: return enumerate(self)

    @property
    def Count(self) -> int: return len(self)
    @property
    def Empty(self) -> bool: return self.Count == 0

    def _Filter(self, func: callable): return filter(func, self)
    def extend(self, items: Union[List[_T], Set[_T]]):
        return self.update(items)
        # for _item in items: self.add(_item)

    def ToList(self) -> List[_T]: return [i for i in self]
    def ToDict(self) -> Dict[int, _T]: return dict(self.enumerate())

    @staticmethod
    def FromArgs(*items: _T): return BaseSetModel.Parse(items)

    @classmethod
    def Parse(cls, d):
        if isinstance(d, list):
            return cls(map(int, d))

        BaseModel.throw(d, list)

class BaseDictModel(dict, BaseModel, Dict[_KT, _VT]):
    def __init__(self, source: dict = None, **kwargs):
        if source is not None: super().__init__(source, **kwargs)
        else: super().__init__(**kwargs)
    def __str__(self): return self.ToString()

    def __delitem__(self, item: Union[_KT, _VT]): return super().__delitem__(item)
    def __contains__(self, item: Union[_KT, _VT]): return super().__contains__(item)
    def __setitem__(self, key: _KT, value: _VT): return super().__setitem__(key, value)
    def __getitem__(self, key: _KT) -> Optional[_VT]:
        try: return super().__getitem__(key)
        except KeyError: return None
    def __iter__(self) -> Iterable[_VT]: return super().__iter__()

    def values(self) -> Iterable[_VT]: return super().values()
    def items(self) -> Iterable[Tuple[_KT, _VT]]: return super().items()
    def keys(self) -> Iterable[_KT]: return super().keys()
    def enumerate(self) -> Iterable[Tuple[int, _T]]: return enumerate(self)

    @property
    def Count(self) -> int: return len(self)
    @property
    def Empty(self) -> bool: return self.Count == 0

    def _Filter(self, func: callable) -> List[_VT]: return list(filter(func, self.values()))

    def ToList(self) -> List[_T]: return list(self.items())



    def ToDict(self) -> Dict[_KT, Union[_VT, Dict, str]]:
        d = { }
        for key, value in self.items():
            if isinstance(value, Enum): d[key] = value.value

            elif isinstance(value, BaseListModel): d[key] = value

            elif isinstance(value, BaseSetModel): d[key] = value.ToList()

            elif isinstance(value, BaseDictModel): d[key] = value.ToDict()

            elif hasattr(value, 'ToList') and callable(value.ToDict): d[key] = value.ToList()

            elif hasattr(value, 'ToDict') and callable(value.ToDict): d[key] = value.ToDict()

            elif hasattr(value, 'ToString') and callable(value.ToString): d[key] = value.ToString()

            else: d[key] = value
        return d
    def ToJsonString(self) -> str: return dumps(self.ToDict(), indent=4, default=self.serialize)  # , cls=JsonEncoder)

    @classmethod
    def Parse(cls, d):
        if isinstance(d, dict):
            return cls(d)

        BaseModel.throw(d, dict)


    # def Iter(self) -> Iterable[int]:
    #     if 0 in self: return range(0, len(self))
    #     return range(1, len(self) + 1)

# class BaseDictNameIdItem(BaseDictModel, IdMixin[str], NameMixin):
#     def split(self) -> Tuple[str, str]: return self.ID, self.Name
#
#     def test(self):
#         _ = self.ID



import re




__all__ = [
        'IsAttributePrivate',
        ]

private_or_special_function_searcher = re.compile(r"(^__\w+$)|(^_\w+$)|(^__\w+__$)")

def IsAttributePrivate(attr_name: str) -> bool: return private_or_special_function_searcher.search(attr_name) is not None



from itertools import count
from typing import Iterator




__all__ = ['AutoCounter']

class AutoCounter(object):
    _counter: Iterator[int]
    _next: callable
    def __init__(self, *, start: int = 0, step: int = 1):
        self._value = start
        self.reset(start=start, step=step)
    def __call__(self, *args, **kwargs) -> int:
        self._value = self._next()
        return self._value
    @property
    def value(self) -> int: return self._value
    def reset(self, *, start: int = 0, step: int = 1):
        self._counter = count(start=start, step=step)
        self._next = self._counter.__next__

    def __str__(self): return str(self._value)
    def __repr__(self): return f'<{self.__class__.__name__}, value: {self._value}>'



import threading
import time
from abc import ABC
from typing import Union




__all__ = ['AutoStartThread', 'AutoStartTargetedThread', 'Wait']

class AutoStartThread(threading.Thread, ABC):
    def __init__(self, *args, Name: str = None, AutoStart: bool = True, Daemon: bool = True, **kwargs):
        if not Name:
            try: Name = self.__class__.__qualname__
            except AttributeError: Name = self.__class__.__name__

        super().__init__(name=Name, args=args, kwargs=kwargs, daemon=Daemon)
        if AutoStart: self.start()
    def run(self): raise NotImplementedError()

class AutoStartTargetedThread(threading.Thread):
    def __init__(self, target: callable, *args, Name: str = None, AutoStart: bool = True, Daemon: bool = True, **kwargs):
        assert (callable(target))
        if not Name:
            try: Name = target.__qualname__
            except AttributeError: Name = target.__name__

        super().__init__(name=Name, target=target, args=args, kwargs=kwargs, daemon=Daemon)
        if AutoStart: self.start()


def Wait(delay: Union[int, float]): time.sleep(delay)



import sys




__all__ = [
        'get_size', 'sizeof',
        ]

def get_size(obj, seen: set = set()):
    """ Recursively finds size of objects """
    size = sys.getsizeof(obj)
    if seen is None: seen = set()

    obj_id = id(obj)
    if obj_id in seen: return 0

    # Important mark as seen *before* entering recursion to gracefully handle self-referential objects
    seen.add(obj_id)
    if isinstance(obj, dict):
        size += sum([get_size(v, seen) for v in obj.values()])
        size += sum([get_size(k, seen) for k in obj.keys()])

    elif hasattr(obj, '__dict__'):
        size += get_size(obj.__dict__, seen)

    elif hasattr(obj, '__iter__') and not isinstance(obj, (str, bytes, bytearray)):
        size += sum([get_size(i, seen) for i in obj])

    return size


def sizeof(obj):
    size = sys.getsizeof(obj)
    if isinstance(obj, dict): return size + sum(map(sizeof, obj.keys())) + sum(map(sizeof, obj.values()))
    if isinstance(obj, (list, tuple, set, frozenset)): return size + sum(map(sizeof, obj))
    return size



import importlib
from typing import List




__all__ = ['ReadFromFile', 'ReadLinesFromFile', 'GetRequirements', 'GetVersion']


def ReadLinesFromFile(path: str) -> List[str]:
    with open(path, "r") as f:
        return f.readlines()

def ReadFromFile(path: str) -> str:
    with open(path, "r") as f:
        return f.read()


def GetVersion(o) -> str:
    if hasattr(o, '__version__'): return o.__version__
    elif hasattr(o, 'version'): return o.version
    elif hasattr(o, 'Version'): return o.Version
    elif hasattr(o, 'VERSION'): return o.VERSION
    else: raise AttributeError(f"can't get version from {o}")


def GetRequirements(path: str) -> List[str]:
    install_requires = []
    for line in ReadLinesFromFile(path):
        line = line.strip('\n')
        if line == 'pillow': package = importlib.import_module('PIL')
        else: package = importlib.import_module(line)
        VERSION = GetVersion(package)
        if VERSION.__class__.__name__ == 'module': VERSION = GetVersion(VERSION)

        install_requires.append(f'{line}>={VERSION}')
    return install_requires



import base64
import hashlib
import json
import os
import pickle




__all__ = [
        'GetFileData', 'GetHashID',
        ]

def GetFileData(path, load: callable, *, Default=None, Check: callable = None, RemoveOnError: bool = False):
    try:
        with open(path, 'rb') as f:
            dat = load(f)
            if Check: Check(dat)
            return dat
    except FileNotFoundError: return Default
    except (pickle.PickleError, pickle.PicklingError, json.JSONDecodeError):
        if RemoveOnError: os.remove(path)
        return Default

def GetHashID(file_path: str, BlockSize: int = 65536) -> str:
    """
    :param file_path: os.path string to the file.
    :param BlockSize: defaults to 64KB
    :return:
    """
    if os.path.isdir(file_path): raise IsADirectoryError('Argument cannot be a directory.')

    _hasher = hashlib.sha1()
    with open(file_path, 'rb') as f:
        buf = f.read(BlockSize)
        while len(buf) > 0:
            _hasher.update(buf)
            buf = f.read(BlockSize)
    return base64.urlsafe_b64encode(_hasher.digest()).decode()



from typing import *




_AnyID = TypeVar("AnyID", str, int)

class _MixinBase:
    get: callable
    __setitem__: callable


class IdMixin(_MixinBase, Generic[_AnyID]):
    __id_key__ = 'ID'
    @property
    def ID(self) -> _AnyID: return self.get(self.__id_key__, None)
    @ID.setter
    def ID(self, v: _AnyID): self[self.__id_key__] = v


class NameMixin(_MixinBase):
    __name_key__ = 'Name'
    @property
    def Name(self) -> str: return self.get(self.__name_key__, None)
    @Name.setter
    def Name(self, v: str): self[self.__name_key__] = v



class AutoNameMixin:
    @property
    def __name__(self) -> str:
        try: return f'{self.__class__.__module__}.{self.__class__.__qualname__}'
        except AttributeError: return f'{self.__class__.__module__}.{self.__class__.__name__}'

class ItemNameMixin:
    __name__: str = None



from datetime import date, datetime, time
from typing import *




__all__ = [
        'DateTime_ToString', 'ToIsoFormat', 'FromIsoFormat', 'now'
        ]

def DateTime_ToString(dt: Optional[Union[datetime, date, time]], _format: str) -> str or None:
    return dt.strftime(_format) if dt is not None else None
def ToIsoFormat(dt: Optional[Union[datetime, date, time]]) -> str:
    return dt.isoformat() if isinstance(dt, (datetime, date, time)) else None
def FromIsoFormat(dt: Type[Union[datetime, date, time]], value: str) -> Optional[Union[datetime, date, time]]:
    return dt.fromisoformat(value) if isinstance(value, str) else None
def now() -> str: return datetime.now().isoformat()




#
# ------------------------------------------------------------------------------
import importlib
import os

from setuptools import setup

from src.BaseExtensions import __author__, __classifiers__, __email__, __license__, __maintainer__, __maintainer_email__, __name__, __short_description__, __url__, __version__




with open(os.path.abspath("PyPiReadme.md"), "r") as f:
    long_description = f.read()

def getVersion(o):
    if hasattr(o, '__version__'): return o.__version__
    elif hasattr(o, 'version'): return o.version
    elif hasattr(o, 'Version'): return o.Version
    elif hasattr(o, 'VERSION'): return o.VERSION
    else: raise AttributeError("can't get version")


with open(os.path.abspath('./requirements.txt'), "r") as f:
    install_requires = []
    for line in f.readlines():
        line = line.strip('\n')
        if line == 'pillow': package = importlib.import_module('PIL')
        else: package = importlib.import_module(line)
        VERSION = getVersion(package)
        if VERSION.__class__.__name__ == 'module': VERSION = getVersion(VERSION)

        install_requires.append(f'{line}>={VERSION}')
print('install_requires', install_requires)

setup(name=__name__,
      version=__version__,
      packages=[__name__, f'{__name__}.Models'],
      url=__url__,
      license=__license__,
      author=__author__,
      author_email=__email__,
      maintainer=__maintainer__,
      maintainer_email=__maintainer_email__,
      description=__short_description__,
      long_description=long_description,
      long_description_content_type="text/markdown",
      install_requires=install_requires,
      classifiers=__classifiers__,
      keywords=f'{__name__} Base Extensions',
      package_dir={ f'{__name__}': f'src/{__name__}' },
      package_data={
              f'{__name__}':         ['*.py'],
              # f'{__name__}.Events':  [f'{__name__}.Events/*.py'],
              # f'{__name__}.Widgets': [f'{__name__}.Widgets/*.py'],
              # f'{__name__}.Misc':    [f'{__name__}.Misc/*.py'],
              },
      )




from types import FunctionType, MethodType
from typing import Union





__all__ = [
        'CalculateOffset', 'RoundFloat',
        'IsMethod', 'IsFunction',
        ]


def RoundFloat(Float: float, Precision: int) -> str:
    """ Rounds the Float to the given Precision and returns It as string. """
    return f"{Float:.{Precision}f}"

def CalculateOffset(starting: Union[int, float], *args: Union[int, float]) -> int:
    """
        Example: WrapLength = ScreenWidth * Widget.Parent.relwidth * Widget.relwidth * offset

    :param starting: starting value (such as width or height)
    :param args: a list of float or integers to be cumulatively multiplied together.
    :return:
    """
    for arg in args:
        if not isinstance(arg, (int, float)): arg = float(arg)
        starting *= arg
    return int(starting)





def IsMethod(o) -> bool:
    """
        Checks if passed object is a method

        https://stackoverflow.com/questions/37455426/advantages-of-using-methodtype-in-python
    :param o: object being checked
    :type o: any
    :return: weather it is a method
    :rtype: bool
    """
    return isinstance(o, MethodType)
def IsFunction(o) -> bool:
    """
        Checks if passed object is a function

        https://stackoverflow.com/questions/37455426/advantages-of-using-methodtype-in-python
    :param o: object being checked
    :type o: any
    :return: weather it is a method
    :rtype: bool
    """
    return isinstance(o, FunctionType)
